wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000094-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000094-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000094-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000094-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000094-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000094-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000109-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000109-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000109-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000109-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000109-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000109-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000125-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000125-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000125-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000125-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000125-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000125-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000211-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000211-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000211-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000211-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000211-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000211-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000240-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000240-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000240-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000240-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000240-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000240-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000241-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000241-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000241-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000241-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000241-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000241-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000250-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000250-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000250-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000250-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000250-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000250-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000251-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000251-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000251-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000251-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000251-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000251-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000256-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000256-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000256-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000256-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000256-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000256-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000259-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000259-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000259-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000259-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000259-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000259-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000273-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000273-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000273-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000273-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000273-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000273-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000287-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000287-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000287-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000287-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000287-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000287-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000297-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000297-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000297-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000297-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000297-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000297-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000307-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000307-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000307-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000307-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000307-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000307-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000308-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000308-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000308-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000308-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000308-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000308-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000727-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000727-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000727-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000727-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000727-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000727-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000745-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000745-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000745-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000745-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000745-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000745-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000752-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000752-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000752-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000752-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000752-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000752-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000756-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000756-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000756-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000756-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000756-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000756-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000994-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000994-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000994-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000994-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000994-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-000994-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001000-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001000-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001000-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001000-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001000-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001000-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001006-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001006-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001006-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001006-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001006-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001006-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001009-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001009-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001009-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001009-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001009-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001009-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001010-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001010-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001010-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001010-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001010-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001010-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001011-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001011-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001011-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001011-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001011-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001011-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001013-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001013-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001013-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001013-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001013-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001013-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001022-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001022-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001022-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001022-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001022-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001022-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001024-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001024-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001024-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001024-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001024-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001024-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001033-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001033-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001033-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001033-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001033-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001033-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001035-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001035-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001035-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001035-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001035-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001035-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001037-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001037-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001037-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001037-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001037-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001037-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001040-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001040-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001040-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001040-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001040-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001040-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001043-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001043-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001043-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001043-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001043-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001043-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001045-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001045-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001045-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001045-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001045-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001045-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001055-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001055-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001055-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001055-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001055-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001055-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001056-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001056-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001056-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001056-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001056-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001056-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001057-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001057-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001057-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001057-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001057-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001057-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001119-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001119-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001119-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001119-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001119-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001119-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001120-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001120-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001120-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001120-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001120-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001120-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001122-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001122-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001122-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001122-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001122-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001122-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001133-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001133-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001133-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001133-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001133-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001133-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001140-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001140-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001140-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001140-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001140-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001140-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001142-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001142-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001142-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001142-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001142-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001142-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001231-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001231-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001231-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001231-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001231-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001231-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001233-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001233-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001233-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001233-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001233-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001233-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001239-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001239-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001239-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001239-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001239-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001239-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001241-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001241-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001241-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001241-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001241-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001241-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001302-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001302-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001302-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001302-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001302-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001302-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001329-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001329-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001329-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001329-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001329-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001329-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001331-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001331-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001331-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001331-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001331-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001331-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001332-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001332-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001332-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001332-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001332-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001332-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001334-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001334-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001334-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001334-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001334-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001334-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001336-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001336-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001336-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001336-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001336-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001336-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001339-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001339-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001339-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001339-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001339-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001339-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001345-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001345-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001345-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001345-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001345-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001345-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001350-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001350-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001350-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001350-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001350-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001350-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001352-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001352-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001352-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001352-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001352-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001352-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001356-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001356-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001356-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001356-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001356-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001356-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001359-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001359-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001359-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001359-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001359-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001359-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001363-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001363-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001363-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001363-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001363-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001363-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001365-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001365-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001365-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001365-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001365-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001365-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001402-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001402-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001402-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001402-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001402-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001402-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001404-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001404-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001404-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001404-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001404-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001404-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001411-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001411-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001411-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001411-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001411-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001411-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001412-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001412-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001412-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001412-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001412-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001412-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001449-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001449-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001449-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001449-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001449-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001449-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001450-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001450-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001450-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001450-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001450-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001450-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001451-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001451-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001451-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001451-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001451-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001451-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001452-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001452-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001452-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001452-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001452-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001452-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001453-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001453-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001453-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001453-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001453-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001453-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001458-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001458-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001458-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001458-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001458-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001458-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001462-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001462-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001462-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001462-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001462-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001462-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001463-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001463-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001463-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001463-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001463-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001463-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001473-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001473-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001473-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001473-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001473-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001473-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001477-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001477-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001477-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001477-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001477-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001477-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001478-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001478-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001478-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001478-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001478-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001478-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001481-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001481-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001481-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001481-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001481-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001481-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001484-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001484-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001484-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001484-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001484-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001484-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001659-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001659-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001659-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001659-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001659-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001659-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001662-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001662-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001662-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001662-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001662-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001662-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001663-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001663-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001663-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001663-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001663-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001663-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001666-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001666-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001666-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001666-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001666-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001666-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001729-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001729-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001729-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001729-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001729-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001729-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001736-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001736-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001736-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001736-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001736-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001736-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001737-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001737-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001737-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001737-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001737-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001737-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001738-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001738-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001738-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001738-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001738-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001738-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001739-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001739-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001739-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001739-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001739-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001739-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001740-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001740-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001740-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001740-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001740-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001740-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001741-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001741-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001741-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001741-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001741-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001741-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001742-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001742-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001742-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001742-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001742-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001742-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001749-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001749-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001749-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001749-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001749-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001749-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001752-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001752-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001752-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001752-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001752-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001752-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001754-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001754-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001754-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001754-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001754-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001754-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001755-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001755-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001755-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001755-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001755-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001755-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001854-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001854-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001854-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001854-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001854-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001854-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001855-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001855-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001855-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001855-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001855-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001855-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001862-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001862-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001862-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001862-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001862-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001862-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001863-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001863-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001863-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001863-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001863-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001863-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001869-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001869-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001869-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001869-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001869-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001869-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001886-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001886-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001886-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001886-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001886-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001886-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001887-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001887-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001887-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001887-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001887-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001887-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001888-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001888-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001888-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001888-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001888-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001888-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001889-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001889-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001889-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001889-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001889-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001889-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001891-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001891-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001891-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001891-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001891-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001891-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001892-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001892-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001892-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001892-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001892-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001892-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001893-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001893-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001893-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001893-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001893-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001893-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001894-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001894-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001894-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001894-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001894-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001894-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001895-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001895-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001895-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001895-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001895-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001895-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001896-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001896-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001896-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001896-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001896-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001896-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001897-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001897-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001897-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001897-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001897-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001897-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001904-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001904-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001904-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001904-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001904-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001904-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001907-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001907-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001907-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001907-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001907-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001907-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001908-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001908-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001908-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001908-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001908-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001908-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001923-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001923-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001923-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001923-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001923-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001923-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001924-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001924-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001924-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001924-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001924-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001924-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001925-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001925-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001925-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001925-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001925-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001925-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001926-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001926-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001926-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001926-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001926-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001926-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001927-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001927-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001927-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001927-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001927-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001927-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001992-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001992-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001992-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001992-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001992-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-001992-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002029-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002029-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002029-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002029-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002029-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002029-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002062-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002062-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002062-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002062-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002062-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002062-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002064-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002064-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002064-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002064-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002064-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002064-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002065-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002065-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002065-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002065-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002065-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002065-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002068-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002068-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002068-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002068-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002068-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002068-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002069-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002069-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002069-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002069-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002069-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002069-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002073-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002073-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002073-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002073-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002073-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002073-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002074-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002074-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002074-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002074-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002074-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002074-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002075-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002075-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002075-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002075-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002075-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002075-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002076-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002076-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002076-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002076-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002076-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002076-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002078-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002078-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002078-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002078-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002078-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002078-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002124-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002124-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002124-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002124-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002124-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002124-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002125-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002125-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002125-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002125-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002125-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002125-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002126-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002126-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002126-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002126-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002126-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002126-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002127-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002127-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002127-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002127-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002127-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002127-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002131-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002131-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002131-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002131-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002131-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002131-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002133-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002133-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002133-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002133-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002133-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002133-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002134-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002134-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002134-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002134-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002134-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002134-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002137-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002137-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002137-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002137-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002137-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002137-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002138-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002138-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002138-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002138-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002138-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002138-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002140-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002140-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002140-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002140-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002140-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002140-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002141-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002141-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002141-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002141-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002141-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002141-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002142-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002142-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002142-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002142-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002142-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002142-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002143-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002143-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002143-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002143-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002143-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002143-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002167-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002167-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002167-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002167-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002167-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002167-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002168-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002168-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002168-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002168-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002168-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002168-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002187-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002187-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002187-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002187-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002187-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002187-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002188-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002188-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002188-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002188-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002188-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002188-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002189-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002189-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002189-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002189-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002189-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002189-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002190-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002190-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002190-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002190-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002190-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002190-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002194-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002194-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002194-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002194-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002194-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002194-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002206-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002206-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002206-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002206-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002206-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002206-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002207-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002207-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002207-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002207-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002207-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002207-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002243-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002243-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002243-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002243-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002243-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002243-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002244-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002244-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002244-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002244-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002244-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002244-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002247-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002247-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002247-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002247-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002247-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002247-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002248-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002248-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002248-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002248-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002248-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002248-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002259-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002259-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002259-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002259-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002259-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002259-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002267-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002267-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002267-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002267-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002267-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002267-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002297-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002297-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002297-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002297-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002297-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002297-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002298-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002298-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002298-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002298-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002298-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002298-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002299-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002299-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002299-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002299-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002299-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002299-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002304-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002304-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002304-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002304-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002304-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002304-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002305-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002305-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002305-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002305-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002305-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002305-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002322-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002322-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002322-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002322-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002322-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002322-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002326-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002326-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002326-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002326-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002326-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002326-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002327-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002327-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002327-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002327-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002327-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002327-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002328-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002328-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002328-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002328-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002328-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002328-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002333-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002333-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002333-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002333-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002333-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002333-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002334-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002334-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002334-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002334-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002334-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002334-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002335-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002335-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002335-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002335-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002335-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002335-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002336-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002336-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002336-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002336-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002336-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002336-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002346-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002346-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002346-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002346-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002346-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002346-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002347-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002347-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002347-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002347-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002347-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002347-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002348-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002348-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002348-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002348-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002348-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002348-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002379-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002379-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002379-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002379-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002379-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002379-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002385-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002385-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002385-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002385-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002385-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002385-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002391-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002391-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002391-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002391-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002391-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002391-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002392-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002392-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002392-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002392-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002392-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002392-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002505-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002505-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002505-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002505-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002505-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002505-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002506-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002506-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002506-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002506-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002506-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002506-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002507-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002507-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002507-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002507-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002507-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002507-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002566-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002566-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002566-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002566-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002566-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002566-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002570-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002570-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002570-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002570-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002570-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002570-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002574-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002574-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002574-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002574-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002574-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002574-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002575-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002575-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002575-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002575-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002575-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002575-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002576-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002576-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002576-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002576-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002576-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002576-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002578-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002578-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002578-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002578-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002578-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002578-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002579-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002579-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002579-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002579-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002579-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002579-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002582-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002582-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002582-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002582-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002582-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002582-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002583-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002583-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002583-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002583-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002583-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002583-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002585-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002585-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002585-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002585-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002585-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002585-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002589-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002589-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002589-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002589-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002589-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002589-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002591-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002591-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002591-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002591-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002591-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002591-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002649-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002649-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002649-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002649-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002649-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002649-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002650-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002650-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002650-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002650-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002650-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002650-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002659-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002659-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002659-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002659-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002659-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002659-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002662-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002662-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002662-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002662-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002662-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002662-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002677-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002677-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002677-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002677-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002677-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002677-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002699-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002699-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002699-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002699-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002699-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002699-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002700-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002700-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002700-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002700-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002700-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002700-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002703-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002703-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002703-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002703-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002703-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002703-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002708-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002708-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002708-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002708-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002708-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002708-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002709-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002709-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002709-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002709-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002709-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002709-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002711-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002711-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002711-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002711-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002711-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002711-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002728-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002728-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002728-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002728-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002728-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002728-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002734-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002734-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002734-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002734-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002734-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002734-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002735-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002735-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002735-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002735-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002735-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002735-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002738-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002738-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002738-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002738-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002738-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002738-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002740-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002740-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002740-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002740-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002740-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002740-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002766-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002766-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002766-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002766-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002766-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002766-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002768-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002768-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002768-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002768-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002768-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002768-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002777-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002777-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002777-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002777-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002777-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002777-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002780-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002780-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002780-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002780-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002780-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002780-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002820-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002820-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002820-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002820-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002820-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002820-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002821-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002821-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002821-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002821-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002821-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002821-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002822-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002822-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002822-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002822-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002822-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002822-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002825-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002825-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002825-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002825-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002825-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002825-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002826-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002826-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002826-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002826-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002826-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002826-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002830-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002830-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002830-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002830-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002830-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002830-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002831-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002831-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002831-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002831-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002831-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002831-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002854-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002854-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002854-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002854-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002854-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002854-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002855-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002855-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002855-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002855-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002855-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002855-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002856-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002856-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002856-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002856-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002856-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002856-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002861-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002861-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002861-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002861-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002861-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002861-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002863-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002863-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002863-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002863-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002863-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002863-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002864-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002864-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002864-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002864-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002864-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002864-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002873-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002873-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002873-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002873-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002873-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002873-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002882-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002882-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002882-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002882-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002882-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002882-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002883-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002883-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002883-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002883-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002883-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002883-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002886-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002886-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002886-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002886-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002886-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002886-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002887-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002887-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002887-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002887-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002887-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002887-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002888-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002888-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002888-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002888-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002888-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002888-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002955-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002955-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002955-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002955-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002955-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002955-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002960-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002960-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002960-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002960-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002960-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002960-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002961-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002961-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002961-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002961-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002961-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002961-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002963-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002963-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002963-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002963-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002963-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002963-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002964-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002964-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002964-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002964-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002964-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002964-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002968-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002968-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002968-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002968-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002968-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002968-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002986-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002986-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002986-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002986-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002986-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002986-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002987-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002987-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002987-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002987-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002987-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002987-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002989-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002989-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002989-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002989-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002989-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-002989-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003015-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003015-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003015-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003015-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003015-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003015-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003031-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003031-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003031-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003031-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003031-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003031-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003058-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003058-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003058-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003058-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003058-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003058-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003059-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003059-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003059-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003059-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003059-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003059-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003063-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003063-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003063-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003063-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003063-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003063-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003064-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003064-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003064-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003064-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003064-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003064-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003098-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003098-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003098-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003098-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003098-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003098-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003103-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003103-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003103-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003103-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003103-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003103-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003104-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003104-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003104-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003104-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003104-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003104-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003130-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003130-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003130-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003130-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003130-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003130-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003170-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003170-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003170-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003170-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003170-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003170-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003177-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003177-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003177-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003177-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003177-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003177-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003180-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003180-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003180-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003180-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003180-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003180-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003183-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003183-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003183-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003183-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003183-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003183-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003184-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003184-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003184-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003184-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003184-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003184-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003185-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003185-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003185-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003185-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003185-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003185-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003187-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003187-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003187-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003187-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003187-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003187-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003204-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003204-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003204-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003204-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003204-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003204-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003206-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003206-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003206-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003206-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003206-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003206-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003225-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003225-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003225-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003225-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003225-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003225-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003226-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003226-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003226-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003226-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003226-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003226-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003229-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003229-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003229-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003229-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003229-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003229-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003230-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003230-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003230-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003230-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003230-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003230-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003256-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003256-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003256-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003256-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003256-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003256-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003313-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003313-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003313-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003313-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003313-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003313-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003322-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003322-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003322-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003322-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003322-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003322-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003323-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003323-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003323-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003323-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003323-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003323-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003325-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003325-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003325-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003325-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003325-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003325-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003354-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003354-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003354-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003354-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003354-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003354-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003355-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003355-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003355-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003355-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003355-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003355-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003356-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003356-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003356-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003356-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003356-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003356-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003358-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003358-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003358-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003358-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003358-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003358-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003360-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003360-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003360-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003360-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003360-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003360-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003362-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003362-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003362-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003362-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003362-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003362-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003366-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003366-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003366-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003366-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003366-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003366-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003367-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003367-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003367-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003367-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003367-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003367-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003368-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003368-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003368-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003368-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003368-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003368-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003384-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003384-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003384-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003384-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003384-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003384-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003388-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003388-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003388-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003388-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003388-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003388-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003427-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003427-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003427-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003427-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003427-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003427-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003430-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003430-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003430-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003430-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003430-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003430-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003434-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003434-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003434-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003434-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003434-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003434-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003437-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003437-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003437-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003437-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003437-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003437-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003438-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003438-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003438-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003438-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003438-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003438-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003458-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003458-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003458-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003458-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003458-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003458-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003460-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003460-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003460-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003460-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003460-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003460-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003461-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003461-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003461-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003461-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003461-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003461-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003462-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003462-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003462-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003462-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003462-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003462-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003465-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003465-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003465-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003465-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003465-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003465-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003478-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003478-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003478-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003478-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003478-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003478-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003511-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003511-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003511-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003511-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003511-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003511-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003512-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003512-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003512-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003512-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003512-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003512-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003513-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003513-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003513-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003513-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003513-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003513-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003518-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003518-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003518-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003518-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003518-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003518-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003525-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003525-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003525-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003525-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003525-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003525-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003530-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003530-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003530-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003530-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003530-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003530-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003538-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003538-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003538-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003538-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003538-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003538-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003557-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003557-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003557-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003557-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003557-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003557-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003559-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003559-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003559-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003559-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003559-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003559-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003560-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003560-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003560-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003560-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003560-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003560-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003562-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003562-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003562-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003562-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003562-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003562-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003563-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003563-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003563-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003563-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003563-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003563-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003565-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003565-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003565-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003565-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003565-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003565-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003600-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003600-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003600-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003600-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003600-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003600-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003605-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003605-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003605-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003605-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003605-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003605-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003606-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003606-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003606-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003606-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003606-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003606-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003610-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003610-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003610-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003610-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003610-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003610-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003614-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003614-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003614-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003614-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003614-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003614-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003615-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003615-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003615-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003615-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003615-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003615-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003624-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003624-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003624-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003624-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003624-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003624-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003628-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003628-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003628-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003628-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003628-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003628-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003629-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003629-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003629-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003629-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003629-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003629-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003630-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003630-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003630-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003630-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003630-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003630-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003631-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003631-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003631-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003631-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003631-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003631-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003634-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003634-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003634-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003634-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003634-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003634-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003635-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003635-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003635-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003635-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003635-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003635-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003636-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003636-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003636-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003636-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003636-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003636-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003642-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003642-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003642-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003642-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003642-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003642-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003643-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003643-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003643-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003643-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003643-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003643-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003644-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003644-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003644-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003644-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003644-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003644-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003646-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003646-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003646-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003646-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003646-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003646-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003647-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003647-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003647-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003647-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003647-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003647-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003648-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003648-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003648-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003648-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003648-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003648-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003649-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003649-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003649-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003649-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003649-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003649-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003650-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003650-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003650-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003650-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003650-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003650-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003664-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003664-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003664-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003664-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003664-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003664-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003697-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003697-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003697-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003697-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003697-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003697-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003698-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003698-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003698-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003698-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003698-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003698-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003699-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003699-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003699-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003699-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003699-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003699-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003704-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003704-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003704-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003704-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003704-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003704-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003705-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003705-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003705-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003705-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003705-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003705-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003712-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003712-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003712-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003712-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003712-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003712-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003716-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003716-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003716-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003716-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003716-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003716-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003723-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003723-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003723-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003723-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003723-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003723-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003804-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003804-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003804-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003804-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003804-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003804-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003805-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003805-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003805-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003805-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003805-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003805-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003813-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003813-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003813-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003813-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003813-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003813-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003818-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003818-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003818-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003818-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003818-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003818-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003829-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003829-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003829-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003829-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003829-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003829-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003830-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003830-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003830-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003830-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003830-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003830-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003835-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003835-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003835-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003835-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003835-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003835-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003836-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003836-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003836-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003836-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003836-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003836-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003838-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003838-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003838-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003838-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003838-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003838-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003840-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003840-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003840-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003840-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003840-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003840-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003841-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003841-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003841-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003841-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003841-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003841-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003842-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003842-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003842-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003842-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003842-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003842-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003893-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003893-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003893-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003893-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003893-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003893-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003894-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003894-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003894-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003894-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003894-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003894-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003900-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003900-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003900-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003900-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003900-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003900-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003903-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003903-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003903-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003903-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003903-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003903-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003905-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003905-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003905-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003905-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003905-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003905-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003909-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003909-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003909-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003909-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003909-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003909-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003910-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003910-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003910-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003910-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003910-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003910-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003911-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003911-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003911-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003911-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003911-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003911-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003918-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003918-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003918-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003918-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003918-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003918-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003919-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003919-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003919-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003919-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003919-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003919-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003925-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003925-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003925-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003925-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003925-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003925-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003926-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003926-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003926-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003926-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003926-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003926-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003927-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003927-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003927-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003927-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003927-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003927-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003957-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003957-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003957-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003957-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003957-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003957-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003958-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003958-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003958-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003958-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003958-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003958-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003959-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003959-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003959-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003959-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003959-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003959-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003964-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003964-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003964-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003964-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003964-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003964-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003965-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003965-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003965-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003965-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003965-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003965-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003967-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003967-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003967-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003967-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003967-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003967-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003970-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003970-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003970-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003970-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003970-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003970-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003971-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003971-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003971-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003971-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003971-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003971-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003972-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003972-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003972-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003972-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003972-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003972-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003983-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003983-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003983-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003983-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003983-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003983-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003992-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003992-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003992-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003992-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003992-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003992-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003996-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003996-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003996-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003996-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003996-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003996-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003997-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003997-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003997-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003997-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003997-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-003997-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004002-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004002-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004002-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004002-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004002-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004002-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004003-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004003-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004003-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004003-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004003-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004003-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004007-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004007-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004007-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004007-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004007-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004007-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004010-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004010-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004010-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004010-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004010-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004010-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004011-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004011-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004011-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004011-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004011-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004011-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004014-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004014-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004014-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004014-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004014-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004014-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004015-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004015-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004015-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004015-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004015-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004015-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004023-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004023-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004023-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004023-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004023-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004023-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004031-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004031-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004031-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004031-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004031-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004031-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004072-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004072-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004072-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004072-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004072-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004072-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004073-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004073-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004073-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004073-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004073-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004073-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004076-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004076-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004076-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004076-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004076-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004076-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004100-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004100-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004100-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004100-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004100-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004100-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004107-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004107-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004107-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004107-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004107-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004107-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004114-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004114-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004114-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004114-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004114-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004114-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004115-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004115-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004115-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004115-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004115-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004115-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004116-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004116-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004116-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004116-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004116-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004116-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004119-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004119-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004119-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004119-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004119-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004119-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004127-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004127-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004127-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004127-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004127-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004127-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004128-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004128-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004128-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004128-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004128-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004128-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004134-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004134-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004134-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004134-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004134-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004134-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004135-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004135-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004135-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004135-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004135-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004135-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004136-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004136-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004136-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004136-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004136-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004136-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004143-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004143-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004143-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004143-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004143-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004143-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004144-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004144-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004144-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004144-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004144-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004144-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004145-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004145-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004145-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004145-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004145-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004145-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004152-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004152-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004152-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004152-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004152-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004152-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004153-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004153-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004153-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004153-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004153-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004153-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004157-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004157-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004157-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004157-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004157-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004157-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004158-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004158-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004158-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004158-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004158-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004158-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004184-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004184-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004184-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004184-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004184-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004184-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004187-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004187-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004187-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004187-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004187-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004187-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004188-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004188-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004188-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004188-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004188-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004188-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004191-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004191-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004191-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004191-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004191-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004191-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004192-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004192-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004192-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004192-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004192-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004192-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004198-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004198-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004198-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004198-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004198-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004198-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004203-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004203-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004203-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004203-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004203-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004203-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004204-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004204-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004204-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004204-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004204-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004204-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004207-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004207-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004207-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004207-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004207-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004207-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004208-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004208-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004208-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004208-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004208-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004208-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004233-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004233-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004233-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004233-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004233-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004233-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004247-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004247-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004247-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004247-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004247-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004247-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004252-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004252-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004252-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004252-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004252-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004252-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004253-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004253-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004253-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004253-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004253-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004253-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004263-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004263-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004263-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004263-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004263-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004263-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004264-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004264-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004264-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004264-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004264-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004264-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004288-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004288-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004288-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004288-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004288-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004288-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004294-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004294-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004294-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004294-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004294-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004294-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004326-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004326-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004326-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004326-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004326-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004326-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004334-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004334-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004334-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004334-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004334-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004334-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004335-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004335-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004335-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004335-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004335-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004335-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004336-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004336-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004336-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004336-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004336-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004336-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004338-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004338-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004338-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004338-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004338-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004338-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004344-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004344-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004344-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004344-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004344-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004344-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004380-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004380-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004380-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004380-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004380-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004380-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004381-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004381-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004381-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004381-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004381-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004381-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004382-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004382-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004382-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004382-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004382-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004382-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004388-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004388-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004388-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004388-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004388-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004388-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004389-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004389-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004389-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004389-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004389-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004389-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004392-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004392-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004392-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004392-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004392-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004392-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004448-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004448-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004448-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004448-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004448-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004448-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004458-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004458-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004458-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004458-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004458-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004458-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004469-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004469-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004469-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004469-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004469-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004469-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004470-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004470-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004470-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004470-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004470-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004470-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004503-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004503-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004503-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004503-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004503-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004503-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004504-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004504-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004504-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004504-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004504-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004504-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004508-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004508-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004508-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004508-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004508-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004508-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004510-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004510-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004510-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004510-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004510-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004510-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004512-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004512-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004512-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004512-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004512-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004512-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004516-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004516-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004516-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004516-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004516-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004516-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004517-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004517-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004517-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004517-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004517-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004517-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004518-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004518-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004518-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004518-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004518-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004518-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004520-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004520-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004520-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004520-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004520-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004520-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004540-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004540-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004540-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004540-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004540-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004540-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004549-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004549-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004549-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004549-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004549-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004549-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004550-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004550-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004550-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004550-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004550-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004550-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004552-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004552-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004552-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004552-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004552-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004552-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004555-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004555-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004555-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004555-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004555-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004555-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004569-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004569-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004569-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004569-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004569-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004569-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004570-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004570-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004570-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004570-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004570-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004570-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004571-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004571-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004571-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004571-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004571-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004571-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004576-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004576-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004576-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004576-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004576-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004576-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004588-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004588-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004588-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004588-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004588-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004588-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004599-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004599-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004599-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004599-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004599-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004599-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004600-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004600-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004600-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004600-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004600-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004600-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004623-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004623-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004623-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004623-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004623-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004623-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004628-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004628-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004628-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004628-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004628-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004628-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004629-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004629-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004629-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004629-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004629-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004629-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004632-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004632-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004632-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004632-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004632-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004632-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004633-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004633-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004633-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004633-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004633-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004633-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004646-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004646-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004646-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004646-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004646-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004646-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004649-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004649-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004649-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004649-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004649-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004649-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004663-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004663-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004663-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004663-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004663-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004663-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004670-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004670-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004670-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004670-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004670-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004670-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004671-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004671-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004671-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004671-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004671-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004671-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004674-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004674-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004674-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004674-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004674-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004674-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004678-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004678-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004678-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004678-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004678-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004678-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004679-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004679-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004679-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004679-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004679-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004679-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004682-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004682-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004682-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004682-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004682-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004682-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004797-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004797-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004797-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004797-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004797-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004797-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004800-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004800-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004800-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004800-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004800-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004800-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004822-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004822-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004822-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004822-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004822-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004822-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004828-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004828-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004828-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004828-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004828-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004828-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004829-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004829-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004829-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004829-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004829-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004829-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004832-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004832-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004832-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004832-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004832-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004832-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004836-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004836-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004836-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004836-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004836-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004836-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004843-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004843-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004843-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004843-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004843-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004843-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004844-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004844-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004844-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004844-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004844-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004844-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004849-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004849-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004849-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004849-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004849-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004849-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004851-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004851-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004851-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004851-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004851-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004851-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004858-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004858-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004858-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004858-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004858-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004858-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004868-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004868-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004868-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004868-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004868-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004868-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004874-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004874-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004874-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004874-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004874-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004874-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004879-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004879-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004879-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004879-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004879-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004879-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004880-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004880-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004880-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004880-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004880-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004880-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004884-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004884-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004884-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004884-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004884-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004884-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004887-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004887-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004887-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004887-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004887-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004887-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004894-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004894-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004894-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004894-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004894-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004894-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004895-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004895-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004895-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004895-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004895-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004895-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004899-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004899-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004899-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004899-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004899-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004899-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004900-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004900-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004900-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004900-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004900-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004900-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004905-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004905-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004905-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004905-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004905-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004905-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004927-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004927-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004927-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004927-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004927-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004927-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004930-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004930-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004930-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004930-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004930-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004930-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004933-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004933-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004933-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004933-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004933-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004933-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004948-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004948-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004948-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004948-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004948-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-004948-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005036-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005036-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005036-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005036-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005036-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005036-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005037-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005037-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005037-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005037-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005037-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005037-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005038-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005038-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005038-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005038-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005038-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005038-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005042-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005042-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005042-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005042-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005042-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005042-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005043-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005043-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005043-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005043-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005043-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005043-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005044-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005044-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005044-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005044-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005044-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005044-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005045-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005045-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005045-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005045-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005045-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005045-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005046-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005046-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005046-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005046-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005046-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005046-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005047-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005047-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005047-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005047-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005047-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005047-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005052-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005052-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005052-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005052-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005052-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005052-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005060-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005060-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005060-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005060-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005060-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005060-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005061-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005061-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005061-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005061-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005061-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005061-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005064-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005064-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005064-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005064-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005064-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005064-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005065-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005065-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005065-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005065-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005065-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005065-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005066-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005066-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005066-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005066-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005066-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005066-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005069-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005069-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005069-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005069-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005069-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005069-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005070-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005070-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005070-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005070-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005070-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005070-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005071-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005071-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005071-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005071-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005071-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005071-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005072-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005072-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005072-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005072-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005072-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005072-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005079-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005079-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005079-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005079-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005079-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005079-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005080-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005080-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005080-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005080-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005080-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005080-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005087-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005087-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005087-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005087-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005087-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005087-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005112-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005112-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005112-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005112-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005112-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005112-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005115-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005115-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005115-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005115-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005115-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005115-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005116-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005116-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005116-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005116-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005116-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005116-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005125-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005125-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005125-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005125-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005125-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005125-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005137-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005137-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005137-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005137-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005137-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005137-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005138-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005138-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005138-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005138-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005138-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005138-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005140-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005140-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005140-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005140-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005140-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005140-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005181-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005181-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005181-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005181-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005181-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005181-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005182-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005182-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005182-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005182-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005182-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005182-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005183-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005183-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005183-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005183-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005183-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005183-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005194-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005194-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005194-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005194-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005194-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005194-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005224-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005224-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005224-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005224-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005224-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005224-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005225-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005225-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005225-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005225-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005225-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005225-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005237-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005237-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005237-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005237-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005237-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005237-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005301-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005301-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005301-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005301-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005301-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005301-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005308-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005308-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005308-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005308-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005308-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005308-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005312-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005312-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005312-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005312-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005312-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005312-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005313-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005313-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005313-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005313-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005313-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005313-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005314-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005314-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005314-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005314-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005314-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005314-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005317-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005317-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005317-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005317-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005317-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005317-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005318-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005318-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005318-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005318-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005318-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005318-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005322-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005322-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005322-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005322-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005322-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005322-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005323-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005323-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005323-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005323-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005323-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005323-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005326-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005326-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005326-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005326-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005326-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005326-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005327-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005327-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005327-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005327-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005327-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005327-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005330-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005330-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005330-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005330-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005330-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005330-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005360-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005360-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005360-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005360-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005360-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005360-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005362-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005362-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005362-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005362-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005362-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005362-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005374-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005374-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005374-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005374-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005374-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005374-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005375-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005375-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005375-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005375-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005375-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005375-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005376-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005376-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005376-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005376-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005376-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005376-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005377-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005377-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005377-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005377-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005377-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005377-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005378-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005378-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005378-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005378-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005378-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005378-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005381-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005381-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005381-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005381-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005381-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005381-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005382-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005382-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005382-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005382-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005382-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005382-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005384-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005384-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005384-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005384-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005384-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005384-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005385-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005385-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005385-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005385-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005385-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005385-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005387-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005387-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005387-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005387-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005387-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005387-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005390-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005390-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005390-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005390-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005390-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005390-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005391-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005391-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005391-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005391-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005391-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005391-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005392-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005392-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005392-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005392-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005392-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005392-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005395-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005395-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005395-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005395-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005395-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005395-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005396-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005396-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005396-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005396-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005396-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005396-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005398-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005398-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005398-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005398-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005398-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005398-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005401-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005401-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005401-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005401-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005401-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005401-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005402-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005402-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005402-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005402-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005402-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005402-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005403-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005403-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005403-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005403-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005403-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005403-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005405-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005405-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005405-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005405-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005405-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005405-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005408-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005408-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005408-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005408-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005408-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005408-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005409-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005409-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005409-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005409-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005409-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005409-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005410-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005410-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005410-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005410-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005410-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005410-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005414-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005414-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005414-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005414-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005414-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005414-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005415-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005415-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005415-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005415-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005415-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005415-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005416-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005416-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005416-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005416-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005416-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005416-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005421-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005421-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005421-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005421-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005421-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005421-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005598-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005598-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005598-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005598-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005598-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005598-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005636-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005636-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005636-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005636-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005636-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005636-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005640-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005640-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005640-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005640-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005640-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005640-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005641-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005641-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005641-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005641-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005641-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005641-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005655-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005655-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005655-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005655-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005655-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005655-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005656-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005656-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005656-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005656-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005656-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005656-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005657-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005657-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005657-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005657-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005657-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005657-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005711-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005711-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005711-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005711-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005711-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005711-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005714-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005714-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005714-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005714-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005714-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005714-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005794-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005794-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005794-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005794-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005794-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005794-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005803-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005803-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005803-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005803-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005803-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005803-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005814-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005814-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005814-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005814-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005814-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005814-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005815-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005815-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005815-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005815-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005815-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005815-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005817-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005817-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005817-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005817-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005817-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005817-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005824-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005824-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005824-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005824-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005824-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005824-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005848-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005848-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005848-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005848-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005848-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005848-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005849-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005849-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005849-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005849-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005849-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005849-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005867-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005867-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005867-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005867-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005867-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005867-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005921-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005921-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005921-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005921-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005921-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005921-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005925-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005925-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005925-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005925-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005925-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005925-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005933-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005933-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005933-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005933-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005933-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005933-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005934-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005934-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005934-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005934-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005934-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005934-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005935-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005935-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005935-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005935-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005935-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005935-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005959-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005959-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005959-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005959-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005959-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005959-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005960-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005960-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005960-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005960-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005960-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005960-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005972-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005972-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005972-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005972-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005972-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005972-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005973-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005973-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005973-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005973-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005973-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005973-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005976-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005976-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005976-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005976-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005976-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005976-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005982-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005982-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005982-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005982-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005982-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005982-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005999-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005999-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005999-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005999-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005999-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-005999-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006003-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006003-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006003-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006003-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006003-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006003-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006004-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006004-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006004-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006004-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006004-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006004-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006005-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006005-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006005-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006005-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006005-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006005-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006073-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006073-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006073-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006073-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006073-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006073-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006074-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006074-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006074-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006074-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006074-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006074-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006075-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006075-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006075-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006075-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006075-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006075-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006103-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006103-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006103-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006103-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006103-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006103-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006104-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006104-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006104-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006104-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006104-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006104-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006105-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006105-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006105-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006105-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006105-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006105-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006118-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006118-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006118-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006118-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006118-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006118-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006121-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006121-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006121-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006121-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006121-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006121-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006122-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006122-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006122-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006122-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006122-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006122-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006123-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006123-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006123-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006123-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006123-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006123-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006160-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006160-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006160-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006160-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006160-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006160-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006161-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006161-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006161-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006161-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006161-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006161-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006162-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006162-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006162-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006162-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006162-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006162-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006166-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006166-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006166-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006166-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006166-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006166-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006168-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006168-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006168-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006168-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006168-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006168-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006169-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006169-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006169-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006169-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006169-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006169-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006174-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006174-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006174-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006174-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006174-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006174-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006176-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006176-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006176-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006176-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006176-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006176-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006177-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006177-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006177-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006177-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006177-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006177-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006182-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006182-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006182-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006182-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006182-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006182-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006207-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006207-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006207-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006207-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006207-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006207-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006212-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006212-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006212-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006212-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006212-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006212-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006282-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006282-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006282-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006282-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006282-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006282-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006354-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006354-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006354-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006354-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006354-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006354-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006361-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006361-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006361-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006361-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006361-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006361-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006366-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006366-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006366-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006366-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006366-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006366-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006371-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006371-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006371-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006371-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006371-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006371-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006372-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006372-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006372-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006372-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006372-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006372-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006413-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006413-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006413-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006413-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006413-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006413-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006426-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006426-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006426-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006426-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006426-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006426-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006469-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006469-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006469-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006469-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006469-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006469-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006475-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006475-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006475-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006475-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006475-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006475-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006481-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006481-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006481-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006481-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006481-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006481-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006485-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006485-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006485-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006485-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006485-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006485-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006505-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006505-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006505-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006505-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006505-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006505-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006509-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006509-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006509-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006509-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006509-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006509-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006514-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006514-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006514-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006514-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006514-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006514-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006519-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006519-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006519-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006519-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006519-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006519-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006572-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006572-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006572-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006572-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006572-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006572-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006573-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006573-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006573-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006573-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006573-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006573-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006574-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006574-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006574-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006574-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006574-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006574-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006581-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006581-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006581-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006581-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006581-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006581-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006585-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006585-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006585-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006585-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006585-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006585-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006591-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006591-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006591-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006591-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006591-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006591-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006597-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006597-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006597-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006597-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006597-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006597-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006605-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006605-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006605-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006605-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006605-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006605-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006708-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006708-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006708-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006708-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006708-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006708-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006709-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006709-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006709-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006709-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006709-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006709-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006727-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006727-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006727-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006727-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006727-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006727-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006728-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006728-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006728-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006728-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006728-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006728-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006749-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006749-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006749-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006749-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006749-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006749-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006751-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006751-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006751-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006751-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006751-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006751-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006752-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006752-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006752-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006752-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006752-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006752-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006753-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006753-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006753-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006753-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006753-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006753-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006781-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006781-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006781-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006781-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006781-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006781-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006782-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006782-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006782-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006782-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006782-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006782-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006793-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006793-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006793-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006793-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006793-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006793-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006794-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006794-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006794-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006794-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006794-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006794-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006850-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006850-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006850-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006850-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006850-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006850-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006895-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006895-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006895-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006895-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006895-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-006895-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007178-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007178-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007178-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007178-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007178-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007178-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007183-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007183-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007183-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007183-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007183-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007183-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007185-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007185-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007185-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007185-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007185-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007185-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007210-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007210-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007210-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007210-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007210-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007210-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007220-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007220-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007220-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007220-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007220-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007220-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007243-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007243-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007243-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007243-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007243-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007243-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007264-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007264-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007264-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007264-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007264-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007264-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007265-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007265-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007265-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007265-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007265-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007265-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007674-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007674-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007674-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007674-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007674-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007674-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007675-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007675-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007675-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007675-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007675-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007675-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007698-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007698-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007698-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007698-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007698-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007698-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007699-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007699-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007699-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007699-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007699-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007699-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007708-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007708-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007708-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007708-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007708-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007708-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007709-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007709-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007709-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007709-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007709-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007709-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007712-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007712-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007712-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007712-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007712-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007712-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007713-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007713-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007713-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007713-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007713-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007713-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007717-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007717-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007717-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007717-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007717-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007717-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007718-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007718-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007718-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007718-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007718-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007718-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007727-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007727-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007727-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007727-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007727-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007727-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007729-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007729-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007729-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007729-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007729-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007729-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007741-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007741-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007741-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007741-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007741-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007741-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007754-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007754-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007754-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007754-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007754-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007754-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007757-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007757-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007757-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007757-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007757-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007757-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007759-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007759-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007759-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007759-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007759-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007759-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007760-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007760-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007760-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007760-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007760-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007760-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007761-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007761-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007761-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007761-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007761-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007761-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007765-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007765-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007765-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007765-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007765-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007765-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007767-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007767-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007767-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007767-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007767-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007767-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007770-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007770-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007770-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007770-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007770-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007770-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007772-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007772-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007772-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007772-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007772-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007772-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007773-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007773-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007773-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007773-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007773-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007773-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007777-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007777-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007777-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007777-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007777-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007777-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007778-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007778-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007778-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007778-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007778-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007778-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007780-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007780-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007780-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007780-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007780-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007780-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007781-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007781-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007781-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007781-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007781-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007781-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007784-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007784-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007784-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007784-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007784-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007784-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007787-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007787-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007787-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007787-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007787-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007787-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007791-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007791-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007791-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007791-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007791-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007791-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007807-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007807-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007807-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007807-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007807-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007807-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007808-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007808-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007808-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007808-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007808-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007808-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007812-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007812-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007812-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007812-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007812-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007812-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007814-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007814-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007814-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007814-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007814-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007814-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007824-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007824-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007824-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007824-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007824-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007824-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007845-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007845-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007845-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007845-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007845-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007845-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007865-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007865-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007865-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007865-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007865-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007865-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007880-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007880-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007880-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007880-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007880-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007880-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007881-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007881-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007881-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007881-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007881-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007881-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007885-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007885-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007885-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007885-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007885-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007885-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007907-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007907-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007907-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007907-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007907-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007907-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007908-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007908-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007908-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007908-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007908-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007908-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007912-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007912-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007912-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007912-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007912-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007912-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007913-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007913-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007913-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007913-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007913-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007913-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007917-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007917-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007917-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007917-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007917-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007917-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007923-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007923-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007923-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007923-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007923-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007923-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007932-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007932-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007932-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007932-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007932-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007932-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007937-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007937-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007937-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007937-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007937-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007937-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007941-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007941-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007941-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007941-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007941-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007941-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007942-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007942-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007942-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007942-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007942-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-007942-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008038-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008038-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008038-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008038-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008038-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008038-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008039-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008039-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008039-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008039-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008039-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008039-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008049-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008049-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008049-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008049-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008049-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008049-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008054-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008054-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008054-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008054-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008054-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008054-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008055-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008055-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008055-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008055-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008055-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008055-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008056-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008056-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008056-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008056-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008056-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008056-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008060-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008060-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008060-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008060-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008060-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008060-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008061-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008061-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008061-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008061-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008061-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008061-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008062-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008062-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008062-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008062-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008062-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008062-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008066-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008066-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008066-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008066-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008066-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008066-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008067-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008067-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008067-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008067-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008067-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008067-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008083-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008083-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008083-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008083-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008083-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008083-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008095-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008095-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008095-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008095-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008095-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008095-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008096-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008096-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008096-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008096-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008096-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008096-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008097-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008097-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008097-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008097-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008097-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008097-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008100-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008100-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008100-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008100-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008100-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008100-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008101-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008101-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008101-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008101-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008101-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008101-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008102-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008102-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008102-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008102-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008102-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008102-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008103-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008103-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008103-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008103-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008103-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008103-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008105-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008105-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008105-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008105-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008105-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008105-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008108-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008108-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008108-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008108-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008108-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008108-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008109-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008109-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008109-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008109-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008109-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008109-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008110-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008110-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008110-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008110-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008110-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008110-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008111-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008111-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008111-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008111-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008111-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008111-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008112-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008112-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008112-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008112-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008112-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008112-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008115-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008115-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008115-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008115-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008115-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008115-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008116-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008116-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008116-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008116-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008116-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008116-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008117-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008117-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008117-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008117-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008117-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008117-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008142-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008142-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008142-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008142-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008142-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008142-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008149-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008149-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008149-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008149-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008149-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008149-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008150-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008150-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008150-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008150-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008150-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008150-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008151-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008151-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008151-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008151-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008151-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008151-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008155-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008155-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008155-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008155-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008155-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008155-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008156-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008156-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008156-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008156-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008156-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008156-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008157-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008157-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008157-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008157-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008157-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008157-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008158-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008158-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008158-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008158-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008158-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008158-6-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008162-1-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008162-2-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008162-3-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008162-4-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008162-5-star.fits.gz
wget -c https://data.sdss.org/sas/dr12/boss/sweeps/dr9/301/calibObj-008162-6-star.fits.gz